import abc
from typing import (
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Tuple,
    TypeVar,
    Union,
    overload,
)

class istr(str): ...

upstr = istr

_S = Union[str, istr]

_T = TypeVar("_T")

_T_co = TypeVar("_T_co", covariant=True)

_D = TypeVar("_D")

class MultiMapping(Mapping[_S, _T_co]):
    @overload
    @abc.abstractmethod
    def getall(self, key: _S) -> List[_T_co]: ...
    @overload
    @abc.abstractmethod
    def getall(self, key: _S, default: _D) -> Union[List[_T_co], _D]: ...
    @overload
    @abc.abstractmethod
    def getone(self, key: _S) -> _T_co: ...
    @overload
    @abc.abstractmethod
    def getone(self, key: _S, default: _D) -> Union[_T_co, _D]: ...

_Arg = Union[Mapping[_S, _T], Dict[_S, _T], MultiMapping[_T], Iterable[Tuple[_S, _T]]]

class MutableMultiMapping(MultiMapping[_T], MutableMapping[_S, _T], Generic[_T]):
    @abc.abstractmethod
    def add(self, key: _S, value: _T) -> None: ...
    @abc.abstractmethod
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    @abc.abstractmethod
    def popone(self, key: _S) -> _T: ...
    @overload
    @abc.abstractmethod
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    @overload
    @abc.abstractmethod
    def popall(self, key: _S) -> List[_T]: ...
    @overload
    @abc.abstractmethod
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

class MultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    def copy(self) -> MultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __setitem__(self, k: _S, v: _T) -> None: ...
    def __delitem__(self, v: _S) -> None: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    def add(self, key: _S, value: _T) -> None: ...
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    @overload
    def popall(self, key: _S) -> List[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

class CIMultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    def copy(self) -> CIMultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __setitem__(self, k: _S, v: _T) -> None: ...
    def __delitem__(self, v: _S) -> None: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    def add(self, key: _S, value: _T) -> None: ...
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    @overload
    def popall(self, key: _S) -> List[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...

class MultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(
        self, arg: Union[MultiMapping[_T], MutableMultiMapping[_T]]
    ) -> None: ...
    def copy(self) -> MultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...

class CIMultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(
        self, arg: Union[MultiMapping[_T], MutableMultiMapping[_T]]
    ) -> None: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> List[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> Union[List[_T], _D]: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> Union[_T, _D]: ...
    def copy(self) -> CIMultiDict[_T]: ...

def getversion(
    md: Union[MultiDict[_T], CIMultiDict[_T], MultiDictProxy[_T], CIMultiDictProxy[_T]]
) -> int: ...



import gc
import sys
from typing import Type

import pytest

from multidict._compat import USE_CYTHON
from multidict._multidict_py import istr as _istr  # noqa: E402

if USE_CYTHON:
    from multidict._multidict import istr  # type: ignore
else:
    from multidict import istr


IMPLEMENTATION = getattr(sys, "implementation")  # to suppress mypy error


class IStrMixin:

    cls = Type[istr]

    def test_ctor(self):
        s = self.cls()
        assert "" == s

    def test_ctor_str(self):
        s = self.cls("aBcD")
        assert "aBcD" == s

    def test_ctor_istr(self):
        s = self.cls("A")
        s2 = self.cls(s)
        assert "A" == s
        assert s == s2

    def test_ctor_buffer(self):
        s = self.cls(b"aBc")
        assert "b'aBc'" == s

    def test_ctor_repr(self):
        s = self.cls(None)
        assert "None" == s

    def test_str(self):
        s = self.cls("aBcD")
        s1 = str(s)
        assert s1 == "aBcD"
        assert type(s1) is str

    def test_eq(self):
        s1 = "Abc"
        s2 = self.cls(s1)
        assert s1 == s2


class TestPyIStr(IStrMixin):
    cls = _istr

    @staticmethod
    def _create_strs():
        _istr("foobarbaz")
        istr2 = _istr()
        _istr(istr2)

    @pytest.mark.skipif(
        IMPLEMENTATION.name != "cpython", reason="PyPy has different GC implementation"
    )
    def test_leak(self):
        gc.collect()
        cnt = len(gc.get_objects())
        for _ in range(10000):
            self._create_strs()

        gc.collect()
        cnt2 = len(gc.get_objects())
        assert abs(cnt - cnt2) < 10  # on PyPy these numbers are not equal


if USE_CYTHON:

    class TestIStr(IStrMixin):
        cls = istr
